# %%

import os
import sys

import geopandas as gpd
import numpy as np
import pandas as pd

from justclust.data.bolzano import cols, conv, read_data, sez_cols, wcols
from justclust.explore.algclusters import explore_models, get_algclusters
from justclust.explore.preprocs import ApplyW, explore_preprocs, get_preprocs
from justclust.paths.bolzano import (
    clfile,
    clscorefile,
    fimpfile,
    prescorefile,
    report_dir,
    selclfile,
)
from justclust.plots import reporting
import justclust.quarto as qmd

# %%
# read and filter/select data
raw = read_data()

# save to a file:
# raw.to_file("raw.geojson", driver="GeoJSON")

# use selected columns
selected = raw.loc[:, cols].dropna()
selected.rename(columns=conv, inplace=True)

# %%
# Remove error observation
mask = selected['Carbon emission building [ton CO2/m²]'] > 8
selected = selected[~ mask]

# %%
# Explore clustering pre-processing

# create a directory if does not exists
os.makedirs(report_dir, exist_ok=True)

# weighted = selected * wcols
apply_weight = ApplyW(wcols)
preprocs = get_preprocs(selected, apply_weight)

# Select a small subset to test the whole workflow
# pkeys = sorted(preprocs.keys())[:10]
# preprocs = {k: preprocs[k] for k in pkeys}

if prescorefile.exists():
    print(f"Reading pre-processed scores from: {prescorefile}")
    pre_scores = pd.read_excel(prescorefile)
else:
    print(
        f"Pre-processed file ({prescorefile}) not found, "
        "start computing the main metrics"
    )
    pre_scores = explore_preprocs(selected, preprocs)
    pre_scores.to_excel(prescorefile)

print("=" * 50)
print(f"BEST PRE-PROCESSING (10/{len(pre_scores)}):")
print(
    pre_scores.sort_values("hopkins", ascending=True)
    .iloc[:10, :]
    .loc[:, ["short_label"]]
)
print("-" * 50)
print("Hopkins statistics of main pre-processing options")
print(pre_scores["hopkins"].describe())

# %%
# select a preprocessing
pre_key = (
    # "s:RobustScaler(quantile_range=(2, 98))|"
    # "w:true|"
    # "d:DictionaryLearning(alpha=0.1, fit_algorithm='cd', n_jobs=-1)"

    "s:Normalizer()|"
    "w:true|"
    "d:DictionaryLearning(alpha=0.1, fit_algorithm='cd', n_jobs=-1)"

    # "s:QuantileTransformer()|w:false|d:DictionaryLearning(alpha=0.1, fit_algorithm='cd', n_jobs=-1)"
)
if pre_key not in preprocs.keys():
    skeys = "\n".join([f"    - {k!r}" for k in sorted(preprocs.keys())])
    print("You must select a pre-processing available from this list:\n" f"{skeys}")
    sys.exit(1)
pre = preprocs[pre_key]

# %%
# explore the clustering space

if clfile.exists():
    print(f"File: {clfile} exists load cluster from the file")
    print("WARNING: clusters read from file, not computed from scratch!")
    geo_out = gpd.read_file(clfile)
    # force index to maintain consistency with raw and selected DF
    geo_out.set_index("SEZ", drop=False, verify_integrity=True, inplace=True)
    print(f"Reading clusters' scores from: {clscorefile}")
    sc = pd.read_excel(clscorefile, header=0, index_col=0)
    print(f"Reading feature importances from: {fimpfile}")
    fi = pd.read_excel(fimpfile, header=0, index_col=0)
else:
    print(f"File: {clfile} does not exist. Start computing the clusters")
    geo_out = raw.loc[:, sez_cols]
    models = get_algclusters(preprocs=pre)

    sc, fi = explore_models(
        data=selected,
        cldf=geo_out,
        rdir=report_dir,
        models=models,
        n_jobs=-1,
        filters=[
            ("n_clusters", (5.0, 12.0)),
            # silhouette_score: The best value is 1 and the worst value is -1.
            # Values near 0 indicate overlapping clusters. Negative values generally
            # indicate that a sample has been assigned to the wrong cluster, as a
            # different cluster is more similar.
            ("silhouette", (0.5, None)),
            # davies_bouldin_score: The minimum score is zero, with lower values
            # indicating better clustering.
            # ("davies bouldin", (None, 0.6)),
            # calinski_harabasz_score: higher Calinski-Harabasz score relates
            # to a model with better defined clusters
            # ("calinski harabasz", (10, None)),
            # mean_el_per_cl: if you need to filter clusters with a mean
            # number of elements
            # ("mean_el_per_cl", (10, None)),
            # % covered by clusters: percentage of elements that have been assigned
            # to a cluster
            ("% covered by clusters", (80, None)),
        ],
    )

    # %%
    # Export the clusters into separate files

    # split file to make them not too heavy
    scs = sc.set_index("label", drop=False, verify_integrity=True).loc[
        sc.selected.values, :
    ]
    algs = scs.groupby("model")
    for alg, adf in algs:
        if len(adf) > 20:
            for k, kdf in adf.groupby("n_clusters"):
                klfile = report_dir / f"clusters_{alg.lower()}_k{k:03d}.gpkg"
                print(f"Saving {klfile}")
                geo_out.loc[:, sez_cols + list(kdf.index)].to_file(
                    klfile, driver="GPKG", index=True
                )

    # save all columns to open all the clusters
    print(f"Saving {clfile}")
    geo_out.to_file(clfile, driver="GPKG", index=True)
    print(f"Saving {clscorefile}")
    sc.to_excel(clscorefile)
    print(f"Saving {fimpfile}")
    fi.to_excel(fimpfile)

# %%
# Select cluster to be selected

# Select the cluster id that need to be compared
# the ID can be taken from the `clscorefile`
sel_clsts = [
    # label
    # "hdbscan__mcs-15_ms-05_m-euclidean_csm-eom",  # k06 @51.6%
    # "hdbscan__mcs-15_ms-05_m-sqeuclidean_csm-eom",  # k07 @50.2%
    # "hdbscan__mcs-10_ms-05_m-sqeuclidean_csm-eom",  # k09 @60.3%
    # "hdbscan__mcs-10_ms-01_m-euclidean_csm-eom",  # k12 @74.5%
    # "hdbscan__mcs-07_ms-05_m-chebyshev_csm-eom",  # k15 @77.5%

    "hdbscan__mcs-10_ms-00_m-euclidean_csm-eom", # k6  96% .91 .17 2653
    # "hdbscan__mcs-12_ms-02_m-euclidean_csm-eom", # k7  91% .85 .27 2338
    # "hdbscan__mcs-02_ms-05_m-euclidean_csm-eom", # k10 96% .87 .25 1674
]

if len(sel_clsts) == 0:
    print(
        f"You must select the clusters to be analyzed from {clscorefile} "
        "using the cluster label as UUID"
    )
    sccols = [
        "label",
        "% covered by clusters",
        "n_clusters",
        "silhouette",
        "davies bouldin",
        "calinski harabasz",
    ]
    # save the default value
    old_max_rows = pd.get_option("display.max_rows")
    pd.set_option("display.max_rows", None)
    print(
        "Available clusters are:\n",
        sc.loc[sc["selected"], sccols].sort_values(
            by=["n_clusters", "silhouette", "davies bouldin", "calinski harabasz"],
            ascending=[True, False, True, False],
        ),
    )
    # save default value back
    pd.set_option("display.max_rows", old_max_rows)
    sys.exit(1)

for clname in sel_clsts:
    if clname not in geo_out.columns:
        raise ValueError(
            f"The selected cluster is: {clname!r} but "
            "it is not available. Select one of the "
            f"following: {geo_out.columns!r}."
        )

# select only the cluster of interest
clsts = geo_out.loc[
    selected.index,
    sel_clsts
    + [
        "geometry",
    ],
]
print(f"Saving selected clusters {clsts}")
clsts.to_file(selclfile, driver="GPKG", index=True)

for clname in sel_clsts:
    print(f"Reporting: {clname}...")
    reporting(
        data=selected.loc[clsts.index],
        clname=clname,
        clsts=clsts,
        cdir=report_dir / clname,
        pal="turbo",
        preprocs=pre,
    )

    qmd.create_qmd_report(
        city = 'bolzano',
        clname= clname,
        cdir = report_dir / clname
        )

# %%
